@@ -30,15 +30,16 @@
 #include "bolt-store.h"
 #include "bolt-str.h"
 #include "bolt-sysfs.h"
+#include "bolt-time.h"
 #include "bolt-udev.h"
 #include "bolt-unix.h"
+#include "bolt-watchdog.h"
 
 #include "bolt-manager.h"
 
 #include <libudev.h>
 #include <string.h>
 
-#define MSEC_PER_USEC 1000LL
 #define PROBING_SETTLE_TIME_MS 2000 /* in milli-seconds */
 
 typedef struct udev_device udev_device;
@@ -52,6 +53,12 @@ static gboolean bolt_manager_initialize (GInitable    *initable,
                                          GCancellable *cancellable,
                                          GError      **error);
 
+static gboolean bolt_manager_store_init (BoltManager *mgr,
+                                         GError     **error);
+
+static void     bolt_manager_store_upgrade (BoltManager *mgr,
+                                            gboolean    *upgraded);
+
 /* internal manager functions */
 static void          manager_sd_notify_status (BoltManager *mgr);
 
@@ -74,6 +81,8 @@ static void          manager_register_domain (BoltManager *mgr,
 static void          manager_deregister_domain (BoltManager *mgr,
                                                 BoltDomain  *domain);
 
+static void          manager_cleanup_stale_domains (BoltManager *mgr);
+
 /* device related functions */
 static gboolean      manager_load_devices (BoltManager *mgr,
                                            GError     **error);
@@ -153,6 +162,10 @@ static void          handle_device_status_changed (BoltDevice  *dev,
                                                    BoltStatus   old,
                                                    BoltManager *mgr);
 
+static void          handle_device_generation_changed (BoltDevice  *dev,
+                                                       GParamSpec  *unused,
+                                                       BoltManager *mgr);
+
 static void          handle_power_state_changed (GObject    *gobject,
                                                  GParamSpec *pspec,
                                                  gpointer    user_data);
@@ -171,7 +184,7 @@ static void          manager_probing_activity (BoltManager *mgr,
                                                gboolean     weak);
 
 /* force powering */
-static BoltPowerGuard *  manager_maybe_power_controller (BoltManager *mgr);
+static BoltGuard *   manager_maybe_power_controller (BoltManager *mgr);
 
 /* config */
 static void          manager_load_user_config (BoltManager *mgr);
@@ -228,6 +241,7 @@ struct _BoltManager
   BoltPower   *power;
   BoltSecurity security;
   BoltAuthMode authmode;
+  guint        generation;
 
   /* policy enforcer */
   BoltBouncer *bouncer;
@@ -242,6 +256,9 @@ struct _BoltManager
   guint      probing_timeout; /* signal id & indicator */
   gint64     probing_tstamp;  /* time stamp of last activity */
   guint      probing_tsettle; /* how long to indicate after the last activity */
+
+  /* watchdog */
+  BoltWatchdog *dog;
 };
 
 enum {
@@ -253,6 +270,7 @@ enum {
   PROP_SECURITY,
   PROP_AUTHMODE,
   PROP_POWERSTATE,
+  PROP_GENERATION,
 
   PROP_LAST,
   PROP_EXPORTED = PROP_VERSION
@@ -286,9 +304,13 @@ bolt_manager_finalize (GObject *object)
   g_ptr_array_free (mgr->devices, TRUE);
   bolt_domain_clear (&mgr->domains);
 
+  g_clear_pointer (&mgr->config, g_key_file_unref);
+
   g_clear_object (&mgr->power);
   g_clear_object (&mgr->bouncer);
 
+  g_clear_object (&mgr->dog);
+
   G_OBJECT_CLASS (bolt_manager_parent_class)->finalize (object);
 }
 
@@ -327,6 +349,10 @@ bolt_manager_get_property (GObject    *object,
       g_value_set_enum (value, bolt_power_get_state (mgr->power));
       break;
 
+    case PROP_GENERATION:
+      g_value_set_uint (value, mgr->generation);
+      break;
+
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
     }
@@ -336,7 +362,6 @@ static void
 bolt_manager_init (BoltManager *mgr)
 {
   mgr->devices = g_ptr_array_new_with_free_func (g_object_unref);
-  mgr->store = bolt_store_new (bolt_get_store_path ());
 
   mgr->probing_roots = g_ptr_array_new_with_free_func (g_free);
   mgr->probing_tsettle = PROBING_SETTLE_TIME_MS; /* milliseconds */
@@ -346,14 +371,6 @@ bolt_manager_init (BoltManager *mgr)
   /* default configuration */
   mgr->policy = BOLT_POLICY_AUTO;
   mgr->authmode = BOLT_AUTH_ENABLED;
-
-  g_signal_connect_object (mgr->store, "device-added",
-                           G_CALLBACK (handle_store_device_added),
-                           mgr, 0);
-
-  g_signal_connect_object (mgr->store, "device-removed",
-                           G_CALLBACK (handle_store_device_removed),
-                           mgr, 0);
 }
 
 static void
@@ -405,6 +422,12 @@ bolt_manager_class_init (BoltManagerClass *klass)
                        G_PARAM_READABLE |
                        G_PARAM_STATIC_STRINGS);
 
+  props[PROP_GENERATION] =
+    g_param_spec_uint ("generation", "Generation", NULL,
+                       0, G_MAXUINT, 0,
+                       G_PARAM_READABLE |
+                       G_PARAM_STATIC_STRINGS);
+
   g_object_class_install_properties (gobject_class, PROP_LAST, props);
 
 
@@ -457,14 +480,20 @@ bolt_manager_initialize (GInitable    *initable,
                          GCancellable *cancellable,
                          GError      **error)
 {
-  g_autoptr(BoltPowerGuard) power = NULL;
+  g_autoptr(BoltGuard) power = NULL;
   BoltManager *mgr;
   struct udev_enumerate *enumerate;
   struct udev_list_entry *l, *devices;
+  gboolean upgraded = FALSE;
   gboolean ok;
 
   mgr = BOLT_MANAGER (initable);
 
+  /* store setup */
+  ok = bolt_manager_store_init (mgr, error);
+  if (!ok)
+    return FALSE;
+
   /* load dynamic user configuration */
   manager_load_user_config (mgr);
 
@@ -475,6 +504,11 @@ bolt_manager_initialize (GInitable    *initable,
 
   bolt_bouncer_add_client (mgr->bouncer, mgr);
 
+  /* watchdog setup */
+  mgr->dog = bolt_watchdog_new (error);
+  if (mgr->dog == NULL)
+    return FALSE;
+
   /* udev setup*/
   bolt_info (LOG_TOPIC ("udev"), "initializing udev");
   mgr->udev = bolt_udev_new ("udev", NULL, error);
@@ -507,7 +541,7 @@ bolt_manager_initialize (GInitable    *initable,
 
   if (power != NULL)
     bolt_info (LOG_TOPIC ("manager"), "acquired power guard '%s'",
-               bolt_power_guard_get_id (power));
+               bolt_guard_get_id (power));
 
   /* TODO: error checking */
   enumerate =  bolt_udev_new_enumerate (mgr->udev, NULL);
@@ -541,19 +575,76 @@ bolt_manager_initialize (GInitable    *initable,
       if (bolt_streq (devtype, "thunderbolt_domain"))
         handle_udev_domain_event (mgr, udevice, "add");
 
-      if (!bolt_streq (devtype, "thunderbolt_device"))
-        continue;
-
-      handle_udev_device_event (mgr, udevice, "add");
+      if (bolt_streq (devtype, "thunderbolt_device"))
+        handle_udev_device_event (mgr, udevice, "add");
     }
 
   udev_enumerate_unref (enumerate);
 
+  /* upgrade the store, if needed */
+  bolt_manager_store_upgrade (mgr, &upgraded);
+
+  if (upgraded)
+    manager_cleanup_stale_domains (mgr);
+
   manager_sd_notify_status (mgr);
 
   return TRUE;
 }
 
+static gboolean
+bolt_manager_store_init (BoltManager *mgr, GError **error)
+{
+  bolt_info (LOG_TOPIC ("manager"), "initializing store");
+
+  mgr->store = bolt_store_new (bolt_get_store_path (), error);
+  if (mgr->store == NULL)
+    return FALSE;
+
+  g_signal_connect_object (mgr->store, "device-added",
+                           G_CALLBACK (handle_store_device_added),
+                           mgr, 0);
+
+  g_signal_connect_object (mgr->store, "device-removed",
+                           G_CALLBACK (handle_store_device_removed),
+                           mgr, 0);
+
+  return TRUE;
+}
+
+static void
+bolt_manager_store_upgrade (BoltManager *mgr, gboolean *upgraded)
+{
+  g_autoptr(GError) err = NULL;
+  BoltStore *store = mgr->store;
+  guint ver;
+  gboolean ok;
+
+  ver = bolt_store_get_version (store);
+
+  if (ver == BOLT_STORE_VERSION)
+    {
+      bolt_debug (LOG_TOPIC ("store"), "store is up to date");
+      return;
+    }
+
+  bolt_info (LOG_TOPIC ("store"), "attempting upgrade from '%d'",
+             ver);
+
+  ok = bolt_store_upgrade (store, NULL, &err);
+  if (!ok)
+    {
+      bolt_warn_err (err, LOG_TOPIC ("store"), "upgrade failed");
+      return;
+    }
+
+  ver = bolt_store_get_version (store);
+  bolt_info (LOG_TOPIC ("store"), "upgraded to version '%d'",
+             ver);
+
+  *upgraded = TRUE;
+}
+
 /* internal functions */
 static void
 manager_sd_notify_status (BoltManager *mgr)
@@ -611,6 +702,24 @@ manager_maybe_set_security (BoltManager *mgr,
     }
 }
 
+static void
+manager_maybe_set_generation (BoltManager *mgr,
+                              guint        gen)
+{
+  guint check;
+
+  check = MAX (mgr->generation, gen);
+
+  if (mgr->generation >= check)
+    return;
+
+  bolt_info ("global 'generation' set to '%d'", check);
+
+  mgr->generation = check;
+  g_object_notify_by_pspec (G_OBJECT (mgr),
+                            props[PROP_GENERATION]);
+}
+
 /* domain related function */
 static gboolean
 manager_load_domains (BoltManager *mgr,
@@ -710,6 +819,58 @@ manager_bootacl_inital_sync (BoltManager *mgr,
              bolt_yesno (ok), empty);
 }
 
+static gboolean
+domain_has_stable_uuid (BoltDomain         *domain,
+                        struct udev_device *dev)
+{
+  g_autoptr(GError) err = NULL;
+  gboolean stable;
+  gboolean ok;
+  guint32 pci_id;
+
+  /* On integrated TBT, like ICL/TGL, the uuid of the
+   * controller is randomly generated on *every* boot,
+   * and thus the uuid is not stable. */
+
+  /* default to FALSE, in case we have no entry */
+  stable = FALSE;
+
+  ok = bolt_sysfs_nhi_id_for_domain (dev, &pci_id, &err);
+  if (!ok)
+    {
+      bolt_warn_err (err, LOG_TOPIC ("udev"), LOG_DOM (domain),
+                     "failed to get NHI for domain");
+      return FALSE;
+    }
+
+  ok = bolt_nhi_uuid_is_stable (pci_id, &stable, &err);
+  if (!ok)
+    bolt_warn_err (err, LOG_TOPIC ("udev"), LOG_DOM (domain),
+                   "failed to determine if uid is stable");
+
+  bolt_info (LOG_TOPIC ("udev"), LOG_DOM (domain),
+             "uuid is stable: %s (for NHI: 0x%04x)",
+             bolt_yesno (stable), pci_id);
+
+  return stable;
+}
+
+static void
+manager_store_domain (BoltManager *mgr,
+                      BoltDomain  *domain)
+{
+  g_autoptr(GError) err = NULL;
+  gboolean ok;
+
+  bolt_info (LOG_TOPIC ("store"), LOG_DOM (domain),
+             "storing newly connected domain");
+
+  ok = bolt_store_put_domain (mgr->store, domain, &err);
+  if (!ok)
+    bolt_warn_err (err, LOG_TOPIC ("store"), LOG_DOM (domain),
+                   "could not store domain");
+}
+
 static BoltDomain *
 manager_domain_ensure (BoltManager        *mgr,
                        struct udev_device *dev)
@@ -725,7 +886,6 @@ manager_domain_ensure (BoltManager        *mgr,
   const char *op;
   const char *uid;
   gboolean iommu;
-  gboolean ok;
 
   /* check if we already know a domain that is the parent
    * of the device (dev); if not then 'dev' is very likely
@@ -746,7 +906,7 @@ manager_domain_ensure (BoltManager        *mgr,
   if (dom == NULL)
     return NULL;
 
-  uid = udev_device_get_sysattr_value (host, "unique_id");
+  uid = bolt_sysfs_device_get_unique_id (host, NULL);
 
   /* check if we have a stored domain with a matching uid
    * of the host device, if so then we have just connected
@@ -767,7 +927,8 @@ manager_domain_ensure (BoltManager        *mgr,
   if (domain == NULL)
     {
       bolt_warn_err (err, LOG_TOPIC ("udev"),
-                     "failed to create domain: %s");
+                     "failed to create domain: %s",
+                     udev_device_get_syspath (dom));
       return NULL;
     }
 
@@ -788,15 +949,12 @@ manager_domain_ensure (BoltManager        *mgr,
   /* add all devices with POLICY_AUTO to the bootacl */
   manager_bootacl_inital_sync (mgr, domain);
 
-  /* now store the domain (with an updated bootacl) */
-  bolt_info (LOG_TOPIC ("store"), LOG_DOM (domain),
-             "storing newly connected domain");
-
-  ok = bolt_store_put_domain (mgr->store, domain, &err);
-  if (!ok)
-    bolt_warn_err (err, LOG_TOPIC ("store"), LOG_DOM (domain),
-                   "could not store domain");
+  /* now store the domain (with an updated bootacl),
+   * but only if its uuid is the same across reboots */
+  if (domain_has_stable_uuid (domain, dom))
+    manager_store_domain (mgr, domain);
 
+  /* export it on the bus and emit the added signals */
   bus = bolt_exported_get_connection (BOLT_EXPORTED (mgr));
   if (bus == NULL)
     return domain;
@@ -904,20 +1062,72 @@ manager_register_domain (BoltManager *mgr,
   g_signal_connect_object (domain, "notify::security",
                            G_CALLBACK (handle_domain_security_changed),
                            mgr, G_CONNECT_SWAPPED);
+
+  bolt_bouncer_add_client (mgr->bouncer, domain);
 }
 
 static void
 manager_deregister_domain (BoltManager *mgr,
                            BoltDomain  *domain)
 {
-  const char *name;
-
-  name = bolt_domain_get_id (domain);
-  bolt_info (LOG_TOPIC ("domain"), "'%s' removed", name);
+  bolt_info (LOG_TOPIC ("manager"), LOG_DOM (domain),
+             "de-registered");
 
   mgr->domains = bolt_domain_remove (mgr->domains, domain);
 }
 
+static void
+manager_cleanup_stale_domains (BoltManager *mgr)
+{
+  g_autoptr(GPtrArray) domains = NULL;
+  BoltDomain *iter;
+  guint count;
+
+  /* Before bolt 0.9.1, domains were stored that have an
+   * unstable uuid, i.e. their uuids change on every boot
+   * and thus there will be store entries that can't be
+   * matched anymore and thus are "stale".
+   */
+
+  bolt_info (LOG_TOPIC ("manager"), "stale domain cleanup");
+
+  count = bolt_domain_count (mgr->domains);
+  domains = g_ptr_array_new_full (count, g_object_unref);
+
+  iter = mgr->domains;
+  for (guint i = 0; i < count; i++)
+    {
+      gboolean stored = bolt_domain_is_stored (iter);
+      gboolean offline = !bolt_domain_is_connected (iter);
+
+      if (stored && offline)
+        g_ptr_array_add (domains, g_object_ref (iter));
+
+      iter = bolt_domain_next (iter);
+    }
+
+  for (guint i = 0; i < domains->len; i++)
+    {
+      g_autoptr(GError) err = NULL;
+      BoltDomain *dom = g_ptr_array_index (domains, i);
+      gboolean ok;
+
+      bolt_info (LOG_DOM (dom), LOG_TOPIC ("store"),
+                 "stale domain detected");
+
+      ok = bolt_store_del_domain (mgr->store, dom, &err);
+
+      if (!ok)
+        {
+          bolt_warn_err (err, LOG_DOM (dom),
+                         "failed to delete domain");
+          continue;
+        }
+
+      manager_deregister_domain (mgr, dom);
+    }
+}
+
 /* device related functions */
 static gboolean
 manager_load_devices (BoltManager *mgr,
@@ -966,13 +1176,38 @@ manager_register_device (BoltManager *mgr,
   g_signal_connect_object (dev, "status-changed",
                            G_CALLBACK (handle_device_status_changed),
                            mgr, 0);
+
+  if (bolt_device_is_host (dev))
+    {
+      guint generation = bolt_device_get_generation (dev);
+      manager_maybe_set_generation (mgr, generation);
+      g_signal_connect_object (dev, "notify::generation",
+                               G_CALLBACK (handle_device_generation_changed),
+                               mgr, 0);
+    }
+
 }
 
 static void
 manager_deregister_device (BoltManager *mgr,
                            BoltDevice  *dev)
 {
+  const char *opath;
+
   g_ptr_array_remove_fast (mgr->devices, dev);
+
+  opath = bolt_device_get_object_path (dev);
+
+  if (opath == NULL)
+    return;
+
+  bolt_exported_emit_signal (BOLT_EXPORTED (mgr),
+                             "DeviceRemoved",
+                             g_variant_new ("(o)", opath),
+                             NULL);
+
+  bolt_device_unexport (dev);
+  bolt_info (LOG_DEV (dev), LOG_TOPIC ("dbus"), "unexported");
 }
 
 static BoltDevice *
@@ -1086,6 +1321,7 @@ bolt_manager_label_device (BoltManager *mgr,
     const char *from;
     const char *to;
   } vendors[] = {
+    {"Dell Inc.",   "Dell" },
     {"HP Inc.",     "HP"   },
     {"Apple, Inc.", "Apple"}
   };
@@ -1123,7 +1359,7 @@ bolt_manager_label_device (BoltManager *mgr,
     }
 
   /* we counted the target too, > 1 means duplicates */
-  if (count > 1)
+  if (count > 1 && !bolt_device_is_host (target))
     label = g_strdup_printf ("%s %s #%u", vendor, name, count);
   else
     label = g_strdup_printf ("%s %s", vendor, name);
@@ -1231,12 +1467,29 @@ manager_auto_authorize (BoltManager *mgr,
   bolt_device_authorize_idle (dev, auth, auto_auth_done, mgr);
 }
 
+static void
+manager_do_import_device (BoltManager *mgr,
+                          BoltDevice  *dev,
+                          BoltPolicy   policy)
+{
+  g_autoptr(GError) err = NULL;
+  gboolean ok;
+
+  ok = bolt_store_put_device (mgr->store,
+                              dev,
+                              policy,
+                              NULL,
+                              &err);
+
+  if (!ok)
+    bolt_warn_err (err, LOG_DEV (dev), LOG_TOPIC ("import"),
+                   "failed to store device");
+}
+
 static void
 manager_maybe_import (BoltManager *mgr,
                       BoltDevice  *dev)
 {
-  g_autoptr(GError) err = NULL;
-  g_autoptr(BoltKey) key = NULL;
   BoltSecurity level;
   BoltPolicy policy;
   const char *secstr;
@@ -1245,18 +1498,36 @@ manager_maybe_import (BoltManager *mgr,
   gboolean boot, pcie;
   gboolean import;
   gboolean iommu;
-  gboolean ok;
-
-  if (bolt_device_get_device_type (dev) == BOLT_DEVICE_HOST)
-    return;
-
-  g_return_if_fail (!bolt_device_get_stored (dev));
-  g_return_if_fail (bolt_device_is_authorized (dev));
 
   /* This function is intentionally more verbose then it
    * could be, but since it is security critical, it is
    * better to be clear than concise */
 
+  g_return_if_fail (!bolt_device_get_stored (dev));
+  g_return_if_fail (bolt_device_is_authorized (dev));
+
+  if (bolt_device_is_host (dev))
+    {
+      BoltDomain *dom = bolt_device_get_domain (dev);
+
+      /* host devices are per design authorized and
+       * must therefore never be authorized by bolt */
+      policy = BOLT_POLICY_MANUAL;
+
+      /* Store the host device only if its domain is
+       * stored as well. Currently, the only reason
+       * for a domain to not be stored is that its
+       * uuid is not stable, i.e. changes every boot.
+       * But its uuid is in fact derived from the
+       * associated host device, i.e. this very host
+       * device here. Ergo, its uuid is unstable and
+       * thus it should not be stored */
+      if (bolt_domain_is_stored (dom))
+        manager_do_import_device (mgr, dev, policy);
+
+      return;
+    }
+
   level = bolt_device_get_security (dev);
   iommu = bolt_device_has_iommu (dev);
   boot = bolt_device_check_authflag (dev, BOLT_AUTH_BOOT);
@@ -1280,22 +1551,12 @@ manager_maybe_import (BoltManager *mgr,
   polstr = bolt_policy_to_string (policy);
 
   bolt_msg (LOG_DEV (dev), LOG_TOPIC ("import"),
-            "%s mode, boot: %s, key: %s -> %s",
-            secstr, bolt_yesno (boot), bolt_yesno (key),
+            "%s mode, boot: %s -> %s",
+            secstr, bolt_yesno (boot),
             (import ? polstr : "no import"));
 
-  if (!import)
-    return;
-
-  ok = bolt_store_put_device (mgr->store,
-                              dev,
-                              policy,
-                              key, &
-                              err);
-
-  if (!ok)
-    bolt_warn_err (err, LOG_DEV (dev), LOG_TOPIC ("import"),
-                   "failed to store device");
+  if (import)
+    manager_do_import_device (mgr, dev, policy);
 }
 
 static void
@@ -1563,7 +1824,7 @@ handle_udev_device_event (BoltManager        *mgr,
 
       /* filter sysfs devices (e.g. the domain) that don't have
        * the unique_id attribute */
-      uid = udev_device_get_sysattr_value (device, "unique_id");
+      uid = bolt_sysfs_device_get_unique_id (device, NULL);
       if (uid == NULL)
         return;
 
@@ -1626,7 +1887,9 @@ handle_udev_device_added (BoltManager        *mgr,
   dev = bolt_device_new_for_udev (udev, domain, &err);
   if (dev == NULL)
     {
-      bolt_warn_err (err, LOG_TOPIC ("udev"), "could not create device");
+      bolt_warn_err (err, LOG_TOPIC ("udev"),
+                     "could not create device for %s",
+                     syspath);
       return;
     }
 
@@ -1687,26 +1950,12 @@ static void
 handle_udev_device_removed (BoltManager *mgr,
                             BoltDevice  *dev)
 {
-  const char *opath;
   const char *syspath;
 
   syspath = bolt_device_get_syspath (dev);
   bolt_msg (LOG_DEV (dev), "removed (%s)", syspath);
 
   manager_deregister_device (mgr, dev);
-
-  opath = bolt_device_get_object_path (dev);
-
-  if (opath == NULL)
-    return;
-
-  bolt_exported_emit_signal (BOLT_EXPORTED (mgr),
-                             "DeviceRemoved",
-                             g_variant_new ("(o)", opath),
-                             NULL);
-
-  bolt_device_unexport (dev);
-  bolt_info (LOG_DEV (dev), LOG_TOPIC ("dbus"), "unexported");
 }
 
 static void
@@ -1759,71 +2008,6 @@ handle_udev_device_detached (BoltManager *mgr,
   bolt_device_disconnected (dev);
 }
 
-typedef struct
-{
-  BoltManager *mgr;
-  const char  *uid;
-} BootaclCtx;
-
-static void
-bootacl_add_dev (gpointer data,
-                 gpointer user_data)
-{
-  g_autoptr(GError) err = NULL;
-  BoltDomain *dom = data;
-  BootaclCtx *ctx = user_data;
-  gboolean ok;
-
-  bolt_info (LOG_TOPIC ("bootacl"), LOG_DOM (dom),
-             LOG_DEV_UID (ctx->uid),
-             "adding newly stored device to bootacl");
-
-  if (!bolt_domain_supports_bootacl (dom))
-    return;
-
-  if (bolt_domain_bootacl_contains (dom, ctx->uid))
-    return;
-
-  ok = bolt_domain_bootacl_add (dom, ctx->uid, &err);
-  if (!ok)
-    {
-      bolt_warn_err (err, LOG_TOPIC ("bootacl"), LOG_DOM (dom),
-                     LOG_DEV_UID (ctx->uid),
-                     "could not add device [%.17s]",
-                     ctx->uid);
-    }
-}
-
-static void
-bootacl_del_dev (gpointer data,
-                 gpointer user_data)
-{
-  g_autoptr(GError) err = NULL;
-  BoltDomain *dom = data;
-  BootaclCtx *ctx = user_data;
-  gboolean ok;
-
-  bolt_info (LOG_TOPIC ("bootacl"), LOG_DOM (dom),
-             LOG_DEV_UID (ctx->uid),
-             "removing forgotten device from bootacl");
-
-  if (!bolt_domain_supports_bootacl (dom))
-    return;
-
-  if (!bolt_domain_bootacl_contains (dom, ctx->uid))
-    return;
-
-  ok = bolt_domain_bootacl_del (dom, ctx->uid, &err);
-  if (!ok)
-    {
-      bolt_warn_err (err, LOG_TOPIC ("bootacl"),
-                     LOG_DEV_UID (ctx->uid), LOG_DOM (dom),
-                     "could not remove device [%.17s]",
-                     ctx->uid);
-    }
-}
-
-
 static void
 handle_store_device_added (BoltStore   *store,
                            const char  *uid,
@@ -1831,7 +2015,6 @@ handle_store_device_added (BoltStore   *store,
 {
   g_autoptr(BoltDevice) dev = NULL;
   BoltDomain *dom = mgr->domains;
-  BootaclCtx ctx = {mgr, uid};
 
   dev = manager_find_device_by_uid (mgr, uid, NULL);
 
@@ -1846,7 +2029,7 @@ handle_store_device_added (BoltStore   *store,
       return;
     }
 
-  bolt_domain_foreach (dom, bootacl_add_dev, &ctx);
+  bolt_domain_foreach (dom, bolt_bootacl_add, dev);
 }
 
 static void
@@ -1856,9 +2039,7 @@ handle_store_device_removed (BoltStore   *store,
 {
   g_autoptr(BoltDevice) dev = NULL;
   BoltDomain *dom = mgr->domains;
-  BootaclCtx ctx = {mgr, uid};
   BoltStatus status;
-  const char *opath;
 
   dev = manager_find_device_by_uid (mgr, uid, NULL);
 
@@ -1875,7 +2056,7 @@ handle_store_device_removed (BoltStore   *store,
                 NULL);
 
   /* remove device from bootacl */
-  bolt_domain_foreach (dom, bootacl_del_dev, &ctx);
+  bolt_domain_foreach (dom, bolt_bootacl_del, dev);
 
   status = bolt_device_get_status (dev);
   /* if the device is connected, keep it around */
@@ -1883,18 +2064,6 @@ handle_store_device_removed (BoltStore   *store,
     return;
 
   manager_deregister_device (mgr, dev);
-  opath = bolt_device_get_object_path (dev);
-
-  if (opath == NULL)
-    return;
-
-  bolt_exported_emit_signal (BOLT_EXPORTED (mgr),
-                             "DeviceRemoved",
-                             g_variant_new ("(o)", opath),
-                             NULL);
-
-  bolt_device_unexport (dev);
-  bolt_info (LOG_DEV (dev), "unexported");
 }
 
 
@@ -1952,6 +2121,21 @@ handle_device_status_changed (BoltDevice  *dev,
     }
 }
 
+static void
+handle_device_generation_changed (BoltDevice  *dev,
+                                  GParamSpec  *unused,
+                                  BoltManager *mgr)
+{
+  guint gen = bolt_device_get_generation (dev);
+
+  bolt_debug (LOG_DEV (dev), LOG_TOPIC ("generation"),
+              "updated to: %u", gen);
+
+  if (!bolt_device_is_host (dev))
+    return;
+
+  manager_maybe_set_generation (mgr, gen);
+}
 
 static void
 handle_power_state_changed (GObject    *gobject,
@@ -1988,7 +2172,7 @@ probing_timeout (gpointer user_data)
 
   /* dt is in microseconds, probing timeout in
    * milli seconds  */
-  timeout = mgr->probing_tsettle * MSEC_PER_USEC;
+  timeout = mgr->probing_tsettle * BOLT_USEC_PER_MSEC;
   if (dt < timeout)
     return G_SOURCE_CONTINUE;
 
@@ -2028,6 +2212,16 @@ device_is_thunderbolt_root (struct udev_device *dev)
          bolt_streq (driver, "thunderbolt");
 }
 
+static gboolean
+device_is_wakeup (struct udev_device *dev)
+{
+  const char *subsys;
+
+  subsys = udev_device_get_subsystem (dev);
+
+  return bolt_streq (subsys, "wakeup");
+}
+
 static gboolean
 probing_add_root (BoltManager        *mgr,
                   struct udev_device *dev)
@@ -2065,6 +2259,12 @@ manager_probing_device_added (BoltManager        *mgr,
   if (syspath == NULL)
     return;
 
+  /* ignore events for wakeup devices which get removed
+   * and added at random time without any connection to
+   * any physical thunderbolt device */
+  if (device_is_wakeup (dev))
+    return;
+
   roots = mgr->probing_roots;
   for (guint i = 0; i < roots->len; i++)
     {
@@ -2136,11 +2336,11 @@ manager_probing_domain_added (BoltManager        *mgr,
   probing_add_root (mgr, p);
 }
 
-static BoltPowerGuard *
+static BoltGuard *
 manager_maybe_power_controller (BoltManager *mgr)
 {
   g_autoptr(GError) err = NULL;
-  BoltPowerGuard *guard = NULL;
+  BoltGuard *guard = NULL;
   gboolean can_force_power;
   int n;
 
@@ -2149,7 +2349,7 @@ manager_maybe_power_controller (BoltManager *mgr)
   if (can_force_power == FALSE)
     return NULL;
 
-  n = bolt_udev_count_domains (mgr->udev, &err);
+  n = bolt_udev_count_hosts (mgr->udev, &err);
   if (n < 0)
     {
       bolt_warn_err (err, LOG_TOPIC ("udev"),
@@ -2175,7 +2375,7 @@ manager_maybe_power_controller (BoltManager *mgr)
   for (int i = 0; i < 25 && n < 1; i++)
     {
       g_usleep (200000); /* 200 000 us = 0.2s */
-      n = bolt_udev_count_domains (mgr->udev, NULL);
+      n = bolt_udev_count_hosts (mgr->udev, NULL);
     }
 
 out:
