@@ -51,6 +51,11 @@
 
 G_DEFINE_AUTOPTR_CLEANUP_FUNC (FILE, fclose);
 
+/* standard open flags for overwriting a file, i.e. close on
+ * exec (3), open in write only mode, truncate before writing,
+ * and create it if it does not yet exist */
+#define BOLT_O_OVERWRITE (O_CLOEXEC | O_WRONLY | O_TRUNC | O_CREAT)
+
 int
 bolt_open (const char *path, int flags, int mode, GError **error)
 {
@@ -318,6 +323,34 @@ bolt_closedir (DIR     *d,
 }
 
 
+gboolean
+bolt_mkdirat (int         dirfd,
+              const char *name,
+              mode_t      mode,
+              GError    **error)
+{
+  int r = 0;
+
+  g_return_val_if_fail (name != NULL, FALSE);
+  g_return_val_if_fail (error == NULL || *error == NULL, FALSE);
+
+  r = mkdirat (dirfd, name, mode);
+
+  if (r < 0)
+    {
+      g_set_error (error,
+                   G_IO_ERROR,
+                   g_io_error_from_errno (errno),
+                   "failed to create directory '%s': %s",
+                   name,
+                   g_strerror (errno));
+      return FALSE;
+    }
+
+  return TRUE;
+}
+
+
 gboolean
 bolt_rmdir (const char *name,
             GError    **error)
@@ -396,6 +429,35 @@ bolt_unlink_at (int         dirfd,
   return TRUE;
 }
 
+gboolean
+bolt_write_file_at (int         dirfd,
+                    const char *name,
+                    const char *data,
+                    gssize      len,
+                    GError    **error)
+{
+  int fd;
+  gboolean ok;
+
+  g_return_val_if_fail (name != NULL, FALSE);
+  g_return_val_if_fail (data != NULL, FALSE);
+  g_return_val_if_fail (error == NULL || *error == NULL, FALSE);
+
+  fd = bolt_openat (dirfd, name, BOLT_O_OVERWRITE, 0666, error);
+
+  if (fd < 0)
+    return FALSE;
+
+  ok = bolt_write_all (fd, data, len, error);
+
+  if (!ok)
+    (void) close (fd);
+  else
+    ok = bolt_close (fd, error);
+
+  return ok;
+}
+
 char *
 bolt_read_value_at (int         dirfd,
                     const char *name,
@@ -499,6 +561,25 @@ retry:
   return n > 0;
 }
 
+gboolean
+bolt_write_int_at (int         dirfd,
+                   const char *name,
+                   gint        val,
+                   GError    **error)
+{
+  char buf[256] = {0, };
+  gsize n;
+
+  g_return_val_if_fail (name != NULL, FALSE);
+  g_return_val_if_fail (error == NULL || *error == NULL, FALSE);
+
+  n = g_snprintf (buf, sizeof (buf), "%d", val);
+
+  g_assert (n < sizeof (buf));
+
+  return bolt_write_file_at (dirfd, name, buf, n, error);
+}
+
 gboolean
 bolt_read_int_at (int         dirfd,
                   const char *name,
@@ -518,6 +599,44 @@ bolt_read_int_at (int         dirfd,
   return bolt_str_parse_as_int (str, val, error);
 }
 
+gboolean
+bolt_write_uint_at (int         dirfd,
+                    const char *name,
+                    guint       val,
+                    GError    **error)
+{
+  char buf[256] = {0, };
+  gsize n;
+
+  g_return_val_if_fail (name != NULL, FALSE);
+  g_return_val_if_fail (error == NULL || *error == NULL, FALSE);
+
+  n = g_snprintf (buf, sizeof (buf), "%u", val);
+
+  g_assert (n < sizeof (buf));
+
+  return bolt_write_file_at (dirfd, name, buf, n, error);
+}
+
+gboolean
+bolt_read_uint_at (int         dirfd,
+                   const char *name,
+                   guint      *val,
+                   GError    **error)
+{
+  g_autofree char *str = NULL;
+
+  g_return_val_if_fail (name != NULL, FALSE);
+  g_return_val_if_fail (error == NULL || *error == NULL, FALSE);
+
+  str = bolt_read_value_at (dirfd, name, error);
+
+  if (str == NULL)
+    return FALSE;
+
+  return bolt_str_parse_as_uint (str, val, error);
+}
+
 gboolean
 bolt_verify_uid (int         dirfd,
                  const char *want,
@@ -536,6 +655,8 @@ bolt_verify_uid (int         dirfd,
 
   if (have == NULL)
     {
+      /* make clang's static analyzer happy */
+      g_return_val_if_fail (err != NULL, FALSE);
       g_set_error (error, BOLT_ERROR, BOLT_ERROR_FAILED,
                    "unique id verification failed: %s",
                    err->message);
@@ -561,26 +682,11 @@ bolt_file_write_all (const char *fn,
                      gssize      n,
                      GError    **error)
 {
-  int fd = -1;
-  gboolean ok;
-
   g_return_val_if_fail (fn != NULL, FALSE);
   g_return_val_if_fail (data != NULL, FALSE);
   g_return_val_if_fail (error == NULL || *error == NULL, FALSE);
 
-  fd = bolt_open (fn, O_CLOEXEC | O_WRONLY | O_TRUNC | O_CREAT, 0666, error);
-
-  if (fd < 0)
-    return FALSE;
-
-  ok = bolt_write_all (fd, data, n, error);
-
-  if (ok)
-    ok = bolt_close (fd, error);
-  else
-    (void) close (fd);
-
-  return ok;
+  return bolt_write_file_at (AT_FDCWD, fn, data, n, error);
 }
 
 gboolean
@@ -789,6 +895,34 @@ bolt_rename (const char *from,
   return FALSE;
 }
 
+gboolean
+bolt_renameat (int         from_dir,
+               const char *from,
+               int         to_dir,
+               const char *to,
+               GError    **error)
+{
+  int code;
+  int r;
+
+  g_return_val_if_fail (from != NULL, FALSE);
+  g_return_val_if_fail (to != NULL, FALSE);
+  g_return_val_if_fail (error == NULL || *error == NULL, FALSE);
+
+  r = renameat (from_dir, from, to_dir, to);
+
+  if (r == 0)
+    return TRUE;
+
+  code = errno;
+  g_set_error (error, G_IO_ERROR,
+               g_io_error_from_errno (code),
+               "could not rename '%s' to '%s': %s",
+               from, to, g_strerror (code));
+
+  return FALSE;
+}
+
 #if !HAVE_FN_COPY_FILE_RANGE
 static loff_t
 copy_file_range (int          fd_in,
@@ -846,6 +980,50 @@ bolt_copy_bytes (int      fd_from,
   return len == 0;
 }
 
+gboolean
+bolt_dir_is_empty (DIR      *dir,
+                   gboolean *empty,
+                   GError  **error)
+{
+  struct dirent *de = NULL;
+  long n;
+
+  g_return_val_if_fail (dir != NULL, FALSE);
+  g_return_val_if_fail (empty != NULL, FALSE);
+  g_return_val_if_fail (error == NULL || *error == NULL, FALSE);
+
+  *empty = TRUE;
+
+  n = telldir (dir);
+  if (n == -1)
+    {
+      int code = errno;
+
+      g_set_error (error, G_IO_ERROR,
+                   g_io_error_from_errno (code),
+                   "telldir(3) failed: %s",
+                   g_strerror (code));
+
+      return FALSE;
+    }
+
+  rewinddir (dir);
+
+  while ((de = readdir (dir)) != NULL)
+    {
+
+      if (!g_strcmp0 (de->d_name, ".") ||
+          !g_strcmp0 (de->d_name, ".."))
+        continue;
+
+      *empty = FALSE;
+      break;
+    }
+
+  seekdir (dir, n);
+
+  return TRUE;
+}
 
 /* auto cleanup helpers */
 void
