@@ -35,7 +35,7 @@ if [[ -x "$(command -v lcov)" ]]; then
 fi
 
 if [[ -x $(command -v pylint) ]]; then
-    pylint tests/test-integration
+    pylint --disable=C0114,C0115,C0116,C0103,C0209,C0301,C0325,R0022,R0913,R0914,R1714,R1725,R1729,R1732,W0201,W0621,W1514 tests/test-integration
 fi
 
 if [[ -x $(command -v codespell) ]]; then
