@@ -21,6 +21,7 @@
 #include "config.h"
 
 #include "bolt-io.h"
+#include "bolt-fs.h"
 #include "bolt-names.h"
 #include "bolt-str.h"
 
@@ -28,10 +29,12 @@
 
 #include <umockdev.h>
 
+#include <gio/gio.h>
 #include <glib/gstdio.h>
 
 #include <errno.h>
 #include <sys/stat.h>
+#include <unistd.h>
 
 typedef struct _MockDevice MockDevice;
 struct _MockDevice
@@ -55,9 +58,11 @@ mock_device_destroy (gpointer data)
 
   g_clear_pointer (&device->devices, g_hash_table_unref);
   g_free (device->path);
+  g_free (device->idstr);
   g_free (device);
 }
 
+
 typedef struct _MockDomain MockDomain;
 
 struct _MockDomain
@@ -69,6 +74,10 @@ struct _MockDomain
 
   gint        serial;
 
+  guint32     nhi_id;
+  char       *nhi_idstr;
+  char       *nhi_path;
+
   MockDevice *host;
 };
 
@@ -79,6 +88,9 @@ mock_domain_destory (gpointer data)
 
   mock_device_destroy (domain->host);
 
+  g_free (domain->nhi_idstr);
+  g_free (domain->nhi_path);
+
   g_free (domain->idstr);
   g_free (domain->path);
   g_free (domain);
@@ -86,13 +98,14 @@ mock_domain_destory (gpointer data)
 
 /* prototypes */
 
-static MockDevice *    mock_sysfs_device_plug (MockSysfs  *ms,
-                                               MockDomain *domain,
-                                               char       *parent,
-                                               MockDevId  *id,
-                                               guint       authorized,
-                                               const char *key,
-                                               gint        boot);
+static MockDevice *    mock_sysfs_device_plug (MockSysfs     *ms,
+                                               MockDomain    *domain,
+                                               char          *parent,
+                                               MockDevId     *id,
+                                               guint          authorized,
+                                               const char    *key,
+                                               gint           boot,
+                                               BoltLinkSpeed *link);
 
 static void            mock_sysfs_device_unplug (MockSysfs  *ms,
                                                  MockDevice *dev);
@@ -109,6 +122,7 @@ struct _MockSysfs
   char       *force_power;
   GHashTable *domains;
   GHashTable *devices;
+  char       *dmi;
 };
 
 
@@ -132,10 +146,17 @@ mock_sysfs_finalize (GObject *object)
 {
   MockSysfs *ms = MOCK_SYSFS (object);
 
-  g_clear_object (&ms->bed);
+  if (ms->dmi)
+    mock_sysfs_dmi_id_remove (ms);
+
+  if (ms->force_power)
+    mock_sysfs_force_power_remove (ms);
+
   g_clear_pointer (&ms->domains, g_hash_table_unref);
   g_clear_pointer (&ms->devices, g_hash_table_unref);
 
+  g_clear_object (&ms->bed);
+
   G_OBJECT_CLASS (mock_sysfs_parent_class)->finalize (object);
 }
 
@@ -154,19 +175,21 @@ mock_sysfs_init (MockSysfs *ms)
   ms->devices = g_hash_table_new (g_str_hash, g_str_equal);
 
   /* udev_enumerate_scan_devices() will return -ENOENT, if
-   * sys/bus or sys/class directories can not be found
+   * sys/bus or sys/class directories can not be found;
+   * fixed in https://github.com/martinpitt/umockdev/commit/5e8296014346
+   * but work with older versions as well
    */
   sys = umockdev_testbed_get_sys_dir (ms->bed);
 
   bus = g_build_filename (sys, "bus", NULL);
   r = g_mkdir (bus, 0744);
 
-  if (r < 0)
+  if (r < 0 && errno != EEXIST)
     g_warning ("could not create %s", bus);
 
   cls = g_build_filename (sys, "class", NULL);
   r = g_mkdir (cls, 0744);
-  if (r < 0)
+  if (r < 0 && errno != EEXIST)
     g_warning ("could not create %s", bus);
 }
 
@@ -214,20 +237,25 @@ mock_sysfs_class_init (MockSysfsClass *klass)
 #define CONST_STRV(...) (char **) (const char *[]){ __VA_ARGS__}
 
 static MockDevice *
-mock_sysfs_device_plug (MockSysfs  *ms,
-                        MockDomain *domain,
-                        char       *parent,
-                        MockDevId  *id,
-                        guint       authorized,
-                        const char *key,
-                        gint        boot)
+mock_sysfs_device_plug (MockSysfs     *ms,
+                        MockDomain    *domain,
+                        char          *parent,
+                        MockDevId     *id,
+                        guint          authorized,
+                        const char    *key,
+                        gint           boot,
+                        BoltLinkSpeed *link)
 {
   g_autofree char *idstr = NULL;
   g_autofree char *vendor_id = NULL;
   g_autofree char *device_id = NULL;
   g_autofree char *authstr = NULL;
   g_autofree char *bootstr = NULL;
-  const char *props[17] = {NULL, };
+  g_autofree char *rx_speed = NULL;
+  g_autofree char *tx_speed = NULL;
+  g_autofree char *rx_lanes = NULL;
+  g_autofree char *tx_lanes = NULL;
+  const char *props[25] = {NULL, };
   MockDevice *device;
   guint serial;
   char *path;
@@ -236,21 +264,43 @@ mock_sysfs_device_plug (MockSysfs  *ms,
   serial = (domain->serial)++;
   idstr = g_strdup_printf ("%u-%u", domain->id, serial);
 
-  vendor_id = g_strdup_printf ("%d", id->vendor_id);
-  device_id = g_strdup_printf ("%d", id->device_id);
   authstr = g_strdup_printf ("%u", authorized);
 
   i = 0;
-  props[i++] = "vendor";
-  props[i++] = vendor_id;
-  props[i++] = "vendor_name";
-  props[i++] = id->vendor_name;
-  props[i++] = "device";
-  props[i++] = device_id;
-  props[i++] = "device_name";
-  props[i++] = id->device_name;
-  props[i++] = "unique_id";
-  props[i++] = id->unique_id;
+
+  if (id->vendor_id)
+    {
+      vendor_id = g_strdup_printf ("%d", id->vendor_id);
+      props[i++] = "vendor";
+      props[i++] = vendor_id;
+    }
+
+  if (id->vendor_name)
+    {
+      props[i++] = "vendor_name";
+      props[i++] = id->vendor_name;
+    }
+
+  if (id->device_id)
+    {
+      device_id = g_strdup_printf ("%d", id->device_id);
+
+      props[i++] = "device";
+      props[i++] = device_id;
+    }
+
+  if (id->device_name)
+    {
+      props[i++] = "device_name";
+      props[i++] = id->device_name;
+    }
+
+  if (id->unique_id)
+    {
+      props[i++] = "unique_id";
+      props[i++] = id->unique_id;
+    }
+
   props[i++] = "authorized";
   props[i++] = authstr;
 
@@ -267,6 +317,23 @@ mock_sysfs_device_plug (MockSysfs  *ms,
       props[i++] = bootstr;
     }
 
+  if (link)
+    {
+      rx_speed = g_strdup_printf ("%u Gb/s\n", link->rx.speed);
+      tx_speed = g_strdup_printf ("%u Gb/s\n", link->tx.speed);
+      rx_lanes = g_strdup_printf ("%u\n", link->rx.lanes);
+      tx_lanes = g_strdup_printf ("%u\n", link->tx.lanes);
+
+      props[i++] = "rx_speed";
+      props[i++] = rx_speed;
+      props[i++] = "tx_speed";
+      props[i++] = tx_speed;
+      props[i++] = "rx_lanes";
+      props[i++] = rx_lanes;
+      props[i++] = "tx_lanes";
+      props[i++] = tx_lanes;
+    }
+
   props[i++] = NULL;
   g_assert (sizeof (props) >= i);
 
@@ -362,6 +429,7 @@ mock_sysfs_force_power_remove (MockSysfs *ms)
 
   umockdev_testbed_uevent (ms->bed, ms->force_power, "remove");
   umockdev_testbed_remove_device (ms->bed, ms->force_power);
+  g_clear_pointer (&ms->force_power, g_free);
 
   return TRUE;
 }
@@ -437,6 +505,57 @@ mock_sysfs_force_power_enabled (MockSysfs *ms)
   return bolt_streq (data, "1");
 }
 
+
+/* dmi */
+const char *
+mock_sysfs_dmi_id_add (MockSysfs  *ms,
+                       const char *sys_vendor,
+                       const char *product_name,
+                       const char *product_version)
+{
+  const char *props[25] = {NULL, };
+  guint i = 0;
+  char *path;
+
+  g_return_val_if_fail (MOCK_IS_SYSFS (ms), NULL);
+  g_return_val_if_fail (ms->dmi == NULL, NULL);
+
+  props[i++] = BOLT_SYSFS_DMI_SYS_VENDOR;
+  props[i++] = sys_vendor;
+
+  props[i++] = BOLT_SYSFS_DMI_PRODUCT_NAME;
+  props[i++] = product_name;
+
+  props[i++] = BOLT_SYSFS_DMI_PRODUCT_VERSION;
+  props[i++] = product_version;
+
+  props[i++] = NULL;
+  g_assert (sizeof (props) >= i);
+
+  path = umockdev_testbed_add_devicev (ms->bed, "dmi", "id",
+                                       NULL,
+                                       (char **) props,
+                                       NULL);
+
+
+  ms->dmi = path;
+
+  return path;
+}
+
+gboolean
+mock_sysfs_dmi_id_remove (MockSysfs *ms)
+{
+  g_return_val_if_fail (MOCK_IS_SYSFS (ms), FALSE);
+  g_return_val_if_fail (ms->dmi != NULL, FALSE);
+
+  umockdev_testbed_uevent (ms->bed, ms->dmi, "remove");
+  umockdev_testbed_remove_device (ms->bed, ms->dmi);
+  g_clear_pointer (&ms->dmi, g_free);
+
+  return TRUE;
+}
+
 /* public methods: domain */
 
 const char *
@@ -445,11 +564,15 @@ mock_sysfs_domain_add (MockSysfs   *ms,
                        ...)
 {
   g_autofree char *acl = NULL;
+  g_autofree char *nhi_pciid = NULL;
+  g_autofree char *nhi_idstr = NULL;
+  g_autofree char *nhi_path = NULL;
   const char *props[7] = {NULL, };
   const char *secstr;
   const char *key;
   MockDomain *domain;
   va_list args;
+  guint32 nhi = 0x15d2;
   char *idstr = NULL;
   char *path;
   guint id;
@@ -483,6 +606,10 @@ mock_sysfs_domain_add (MockSysfs   *ms,
           props[i++] = BOLT_SYSFS_IOMMU;
           props[i++] = iommu;
         }
+      else if (bolt_streq (key, "nhi"))
+        {
+          nhi = va_arg (args, int);
+        }
     }
 
   g_assert (i < 7);
@@ -492,18 +619,40 @@ mock_sysfs_domain_add (MockSysfs   *ms,
 
   g_assert (sizeof (props) >= i);
 
+  /* native host interface (NHI) */
+  nhi_pciid = g_strdup_printf ("0x%04x", nhi);
+  nhi_idstr = g_strdup_printf ("0000:00:01.%u", id);
+
+  nhi_path = umockdev_testbed_add_devicev (ms->bed, "pci", nhi_idstr,
+                                           NULL, /* parent: NHI has none */
+                                           CONST_STRV ("class", "0x088000",
+                                                       "vendor", "0x8086", /* Intel */
+                                                       "device", nhi_pciid,
+                                                       NULL),
+                                           CONST_STRV ("DRIVER", "thunderbolt", NULL));
+
+  g_debug ("M [A] %s (0x%04x) @ %s", nhi_idstr, nhi, nhi_path);
+
+  /* add the domain */
   path = umockdev_testbed_add_devicev (ms->bed, "thunderbolt", idstr,
-                                       NULL, /* parent: domain has none */
+                                       nhi_path, /* parent */
                                        (char **) props,
                                        CONST_STRV ("DEVTYPE", "thunderbolt_domain", NULL));
 
   if (path == NULL)
-    return path;
+    {
+      umockdev_testbed_remove_device (ms->bed, nhi_path);
+      return path;
+    }
 
   g_debug ("M [A] %s (%s) @ %s", idstr, secstr, path);
 
   domain = g_new0 (MockDomain, 1);
 
+  domain->nhi_id = nhi;
+  domain->nhi_idstr = g_steal_pointer (&nhi_idstr);
+  domain->nhi_path = g_steal_pointer (&nhi_path);
+
   domain->id = id;
   domain->idstr = idstr;
   domain->path = path;
@@ -555,6 +704,11 @@ mock_sysfs_domain_remove (MockSysfs  *ms,
   umockdev_testbed_uevent (ms->bed, domain->path, "remove");
   umockdev_testbed_remove_device (ms->bed, domain->path);
 
+  g_debug ("M [R] %s @ %s", domain->nhi_idstr, domain->nhi_path);
+
+  umockdev_testbed_uevent (ms->bed, domain->nhi_path, "remove");
+  umockdev_testbed_remove_device (ms->bed, domain->nhi_path);
+
   g_hash_table_remove (ms->domains, id);
   return TRUE;
 }
@@ -691,20 +845,65 @@ mock_sysfs_host_add (MockSysfs  *ms,
                                    domain->path,
                                    id,
                                    1,
-                                   NULL, /* no key for the host */
-                                   -1);  /* no boot file either */
+                                   NULL,  /* no key for the host */
+                                   -1,    /* no boot file either */
+                                   NULL); /* no link settings */
   domain->host = device;
 
   return device->idstr;
 }
 
+void
+mock_sysfs_host_remove (MockSysfs  *ms,
+                        const char *host)
+{
+  GHashTableIter iter;
+  gpointer k, v;
+  MockDevice *dev;
+  MockDomain *domain = NULL;
+
+  g_return_if_fail (MOCK_IS_SYSFS (ms));
+  g_return_if_fail (host != NULL);
+
+  dev = g_hash_table_lookup (ms->devices, host);
+
+  if (dev == NULL)
+    {
+      g_error ("Device not found for %s", host);
+      return;
+    }
+
+  g_hash_table_iter_init (&iter, ms->domains);
+  while (g_hash_table_iter_next (&iter, &k, &v))
+    {
+      MockDomain *d = v;
+      if (d->host == dev)
+        {
+          domain = d;
+          break;
+        }
+    }
+
+  if (!domain)
+    {
+      g_error ("domain not found for host: %s", host);
+      return;
+    }
+
+  mock_sysfs_device_unplug (ms, dev);
+  mock_device_destroy (dev);
+  domain->host = NULL;
+}
+
+
 const char *
-mock_sysfs_device_add (MockSysfs  *ms,
-                       const char *parent,
-                       MockDevId  *id,
-                       guint       authorized,
-                       const char *key,
-                       gint        boot)
+mock_sysfs_device_add (MockSysfs     *ms,
+                       const char    *parent,
+                       MockDevId     *id,
+                       guint          authorized,
+                       const char    *key,
+                       gint           boot,
+                       BoltLinkSpeed *speed)
 {
   MockDevice *pdev;
   MockDomain *domain = NULL;
@@ -747,7 +946,8 @@ mock_sysfs_device_add (MockSysfs  *ms,
                                    id,
                                    1,
                                    key,
-                                   boot);
+                                   boot,
+                                   speed);
 
   g_hash_table_insert (pdev->devices, device->idstr, device);
 
@@ -835,3 +1035,65 @@ mock_sysfs_device_remove (MockSysfs  *ms,
 
   return TRUE;
 }
+
+gboolean
+mock_sysfs_set_osrelease (MockSysfs  *ms,
+                          const char *version)
+{
+  g_autoptr(GError) err = NULL;
+  g_autoptr(GFile) target = NULL;
+  g_autofree char *data = NULL;
+  g_autofree char *path = NULL;
+  g_autofree char *root = NULL;
+  gboolean ok;
+  gsize n;
+  int r;
+
+  root = umockdev_testbed_get_root_dir (ms->bed);
+  target = g_file_new_build_filename (root, "proc/sys/kernel/osrelease", NULL);
+
+  ok = bolt_fs_make_parent_dirs (target, &err);
+  if (!ok)
+    {
+      g_debug ("Failed to make parent dirs: %s", err->message);
+      return FALSE;
+    }
+
+  if (version != NULL)
+    data = g_strdup_printf ("%s\n", version);
+  else
+    data = g_strdup ("<broken>\n");
+
+  n = strlen (data);
+
+  /* make sure we can write the file, if it fails,
+   * we will indirectly catch it in the write */
+  path = g_file_get_path (target);
+  r = chmod (path, 0644);
+
+  if (r != 0)
+    g_debug ("Failed to set mode: %s", g_strerror (errno));
+
+  ok = g_file_replace_contents (target,
+                                data, n,
+                                NULL,
+                                FALSE,
+                                0,
+                                NULL,
+                                NULL,
+                                &err);
+
+  if (!ok)
+    g_debug ("Failed to set osrelease: %s", err->message);
+
+  /* make the file non-readable if version == NULL,
+   * so to simulate read errors */
+  if (version == NULL)
+    {
+      r = chmod (path, 0000);
+      if (r != 0)
+        g_debug ("Failed to set mode: %s", g_strerror (errno));
+    }
+
+  return ok;
+}
