@@ -47,13 +47,50 @@ typedef struct
 {
   char      *path;
   BoltStore *store;
+
 } TestStore;
 
+typedef struct
+{
+  GLogWriterFunc logger;
+} TestContext;
+
+static void
+test_context_set_logger (TestContext *data, GLogWriterFunc logger)
+{
+  data->logger = logger;
+}
+
+static GLogWriterOutput
+test_context_logger (GLogLevelFlags   log_level,
+                     const GLogField *fields,
+                     gsize            n_fields,
+                     gpointer         user_data)
+{
+  const TestContext *data = user_data;
+
+  return data->logger (log_level, fields, n_fields, NULL);
+}
+
+static GLogWriterOutput
+null_logger (GLogLevelFlags   log_level,
+             const GLogField *fields,
+             gsize            n_fields,
+             gpointer         user_data)
+{
+  return G_LOG_WRITER_HANDLED;
+}
 
 static void
 test_store_setup (TestStore *tt, gconstpointer user_data)
 {
   g_autoptr(GError) error = NULL;
+  TestContext *ctx = (TestContext *) user_data;
+
+  /* reset logger */
+  test_context_set_logger (ctx, g_log_writer_default);
+
+  ctx->logger = g_log_writer_default;
 
   tt->path = g_dir_make_tmp ("bolt.auth.XXXXXX",
                              &error);
@@ -66,16 +103,15 @@ test_store_setup (TestStore *tt, gconstpointer user_data)
     }
 
 
-  tt->store = bolt_store_new (tt->path);
+  tt->store = bolt_store_new (tt->path, &error);
   if (tt->store == NULL)
     {
-      g_critical ("Could not create store at %s",
-                  tt->path);
+      g_critical ("Could not create store at %s: %s",
+                  tt->path, error->message);
       return;
     }
 
   g_debug ("store at '%s'", tt->path);
-
 }
 
 static void
@@ -105,12 +141,19 @@ test_store_basic (TestStore *tt, gconstpointer user_data)
   g_autofree char *path = NULL;
   char uid[] = "fbc83890-e9bf-45e5-a777-b3728490989c";
   BoltKeyState keystate;
+  guint version;
   gboolean ok;
 
   g_object_get (tt->store, "root", &root, NULL);
   path = g_file_get_path (root);
   g_assert_cmpstr (tt->path, ==, path);
 
+  g_object_get (tt->store, "version", &version, NULL);
+  g_assert_cmpuint (version, ==, BOLT_STORE_VERSION);
+
+  version = bolt_store_get_version (tt->store);
+  g_assert_cmpuint (version, ==, BOLT_STORE_VERSION);
+
   dev = g_object_new (BOLT_TYPE_DEVICE,
                       "uid", uid,
                       "name", "Laptop",
@@ -138,6 +181,7 @@ test_store_basic (TestStore *tt, gconstpointer user_data)
   g_assert_cmpstr (bolt_device_get_uid (stored), ==, bolt_device_get_uid (dev));
   g_assert_cmpstr (bolt_device_get_name (stored), ==, bolt_device_get_name (dev));
   g_assert_cmpstr (bolt_device_get_vendor (stored), ==, bolt_device_get_vendor (dev));
+  g_assert_cmpuint (bolt_device_get_generation (stored), ==, 0);
 
   g_assert_cmpuint (bolt_device_get_policy (stored), ==, BOLT_POLICY_AUTO);
   g_assert_cmpuint (bolt_device_get_stored (stored), ==, TRUE);
@@ -151,6 +195,7 @@ test_store_basic (TestStore *tt, gconstpointer user_data)
                       "uid", uid,
                       "name", "Laptop",
                       "vendor", "GNOME.org",
+                      "generation", 4,
                       "status", BOLT_STATUS_DISCONNECTED,
                       NULL);
 
@@ -176,6 +221,7 @@ test_store_basic (TestStore *tt, gconstpointer user_data)
   g_assert_cmpstr (bolt_device_get_name (stored), ==, bolt_device_get_name (dev));
   g_assert_cmpstr (bolt_device_get_vendor (stored), ==, bolt_device_get_vendor (dev));
 
+  g_assert_cmpuint (bolt_device_get_generation (stored), ==, 4);
   g_assert_cmpuint (bolt_device_get_policy (stored), ==, BOLT_POLICY_MANUAL);
   g_assert_cmpuint (bolt_device_get_stored (stored), ==, TRUE);
   g_assert_cmpuint (bolt_device_get_keystate (stored), ==, 1);
@@ -238,6 +284,69 @@ test_store_basic (TestStore *tt, gconstpointer user_data)
 
 }
 
+static void
+test_store_update (TestStore *tt, gconstpointer user_data)
+{
+  g_autoptr(BoltDevice) dev = NULL;
+  g_autoptr(BoltKey) key = NULL;
+  g_autoptr(GError) err = NULL;
+  char uid[] = "fbc83890-e9bf-45e5-a777-b3728490989c";
+  BoltKeyState keystate;
+  BoltPolicy policy;
+  gboolean ok;
+  guint64 storetime;
+
+  policy = BOLT_POLICY_IOMMU;
+  dev = g_object_new (BOLT_TYPE_DEVICE,
+                      "uid", uid,
+                      "name", "Laptop",
+                      "vendor", "GNOME.org",
+                      "status", BOLT_STATUS_DISCONNECTED,
+                      "generation", 1,
+                      NULL);
+
+  key = bolt_key_new (NULL);
+  g_assert_nonnull (key);
+
+  ok = bolt_store_put_device (tt->store,
+                              dev,
+                              policy,
+                              key,
+                              &err);
+
+  g_assert_no_error (err);
+  g_assert_true (ok);
+
+  keystate = bolt_device_get_keystate (dev);
+  g_assert_cmpuint (keystate, ==, BOLT_KEY_NEW);
+  g_assert_cmpuint (bolt_device_get_policy (dev), ==, policy);
+
+  storetime = bolt_device_get_storetime (dev);
+
+  g_object_set (G_OBJECT (dev),
+                "generation", 3,
+                "label", "My Laptop",
+                NULL);
+
+  /* update the device. generation and label should
+   * change, but the rest should stay the same, esp.
+   * keystate and also storetime should not change.
+   * Also, BOLT_POLICY_DEFAULT should be ignored */
+  ok = bolt_store_put_device (tt->store,
+                              dev,
+                              BOLT_POLICY_DEFAULT,
+                              NULL,
+                              &err);
+
+  g_assert_no_error (err);
+  g_assert_true (ok);
+
+  keystate = bolt_device_get_keystate (dev);
+  g_assert_cmpuint (keystate, ==, BOLT_KEY_NEW);
+  g_assert_cmpuint (bolt_device_get_policy (dev), ==, policy);
+  g_assert_cmpuint (bolt_device_get_storetime (dev), ==, storetime);
+}
+
 static void
 test_store_config (TestStore *tt, gconstpointer user_data)
 {
@@ -345,6 +454,8 @@ test_key (TestStore *tt, gconstpointer user_data)
   g_assert_no_error (err);
   g_assert_nonnull (loaded);
 
+  g_clear_object (&loaded);
+
   /* corrupt the key */
   p = g_file_get_path (f);
   r = truncate (p, 32);
@@ -369,16 +480,6 @@ test_key (TestStore *tt, gconstpointer user_data)
     }
 }
 
-static GLogWriterOutput
-null_logger (GLogLevelFlags   log_level,
-             const GLogField *fields,
-             gsize            n_fields,
-             gpointer         user_data)
-{
-  return G_LOG_WRITER_HANDLED;
-}
-
-
 static void
 test_store_invalid_data (TestStore *tt, gconstpointer user_data)
 {
@@ -388,6 +489,7 @@ test_store_invalid_data (TestStore *tt, gconstpointer user_data)
   g_autoptr(GError) err = NULL;
   g_autoptr(BoltDevice) dev = NULL;
   static const char *uid = "399d33cb-c9cf-4273-8f92-9445437e0b43";
+  TestContext *ctx = (TestContext *) user_data;
   gboolean ok;
   int r;
 
@@ -400,9 +502,9 @@ test_store_invalid_data (TestStore *tt, gconstpointer user_data)
   g_assert_no_error (err);
   g_assert_true (ok);
 
-  g_log_set_writer_func (null_logger, NULL, NULL);
+  test_context_set_logger (ctx, null_logger);
   dev = bolt_store_get_device (tt->store, uid, &err);
-  g_log_set_writer_func (g_log_writer_default, NULL, NULL);
+  test_context_set_logger (ctx, g_log_writer_default);
 
   g_assert_null (dev);
   g_assert_error (err, BOLT_ERROR, BOLT_ERROR_FAILED);
@@ -595,6 +697,7 @@ test_store_domain (TestStore *tt, gconstpointer user_data)
   g_autoptr(BoltDomain) d1 = NULL;
   g_autoptr(BoltDomain) s1 = NULL;
   const char *uid = "884c6edd-7118-4b21-b186-b02d396ecca0";
+  gboolean exists;
   gboolean ok;
   GStrv bootacl = NULL;
   const char *acl[16] = {
@@ -655,6 +758,13 @@ test_store_domain (TestStore *tt, gconstpointer user_data)
   g_assert_no_error (err);
   g_assert_true (ok);
 
+  exists = bolt_store_has_journal (tt->store, "bootacl", uid);
+  g_assert_true (exists);
+
+  ok = bolt_domain_can_delete (d1, &err);
+  g_assert_no_error (err);
+  g_assert_true (ok);
+
   /* update: get again after update */
   g_clear_object (&s1);
   s1 = bolt_store_get_domain (tt->store, uid, &err);
@@ -665,66 +775,235 @@ test_store_domain (TestStore *tt, gconstpointer user_data)
 
   bootacl = bolt_domain_get_bootacl (s1);
   bolt_assert_strv_equal ((GStrv) acl, bootacl, 0);
+  g_clear_object (&s1);
 
   /* delete */
-  g_assert_true (bolt_domain_is_stored (s1));
-  ok = bolt_store_del_domain (tt->store, s1, &err);
+  g_assert_true (bolt_domain_is_stored (d1));
+  ok = bolt_store_del_domain (tt->store, d1, &err);
   g_assert_no_error (err);
   g_assert_true (ok);
-  g_assert_false (bolt_domain_is_stored (s1));
+  g_assert_false (bolt_domain_is_stored (d1));
+
+  exists = bolt_store_has_journal (tt->store, "bootacl", uid);
+  g_assert_false (exists);
+
+  /* store again, modify the bootacl, i.e. write to the journal */
+  ok = bolt_store_put_domain (tt->store, d1, &err);
+  g_assert_no_error (err);
+  g_assert_true (ok);
+  g_assert_true (bolt_domain_is_stored (d1));
+
+  ok = bolt_domain_bootacl_del (d1,
+                                "884c6edd-7118-4b21-b186-b02d396ecca1",
+                                &err);
+  g_assert_no_error (err);
+  g_assert_true (ok);
+
+  bootacl = bolt_domain_get_bootacl (d1);
+  acl[0] = "";
+  bolt_assert_strv_equal ((GStrv) acl, bootacl, 0);
+
+  /* journal should exist and non-empty */
+  exists = bolt_store_has_journal (tt->store, "bootacl", uid);
+  g_assert_true (exists);
+
+  /* non-empty journal should prevent deletion */
+  ok = bolt_domain_can_delete (d1, &err);
+  g_assert_error (err, G_IO_ERROR, G_IO_ERROR_NOT_EMPTY);
+  g_assert_false (ok);
+  g_clear_error (&err);
+
+  ok = bolt_store_del_domain (tt->store, d1, &err);
+  g_assert_error (err, G_IO_ERROR, G_IO_ERROR_NOT_EMPTY);
+  g_assert_false (ok);
+}
+
+static void
+test_store_journal (TestStore *tt, gconstpointer user_data)
+{
+  g_autoptr(GError) err = NULL;
+  g_autoptr(BoltJournal) journal = NULL;
+  gboolean exists;
+  gboolean ok;
+
+  /* delete an non-existing one */
+  ok = bolt_store_del_journal (tt->store, "acl", "log", &err);
+  g_assert_no_error (err);
+  g_assert_true (ok);
+
+  /* create a non-existing one */
+  journal = bolt_store_open_journal (tt->store, "acl", "log", &err);
+  g_assert_no_error (err);
+  g_assert_nonnull (journal);
+
+  exists = bolt_store_has_journal (tt->store, "acl", "log");
+  g_assert_true (exists);
+
+  /* re-open the existing one */
+  g_clear_object (&journal);
+  journal = bolt_store_open_journal (tt->store, "acl", "log", &err);
+  g_assert_no_error (err);
+  g_assert_nonnull (journal);
+
+  /* delete the journal */
+  g_clear_object (&journal);
+  ok = bolt_store_del_journal (tt->store, "acl", "log", &err);
+  g_assert_no_error (err);
+  g_assert_true (ok);
+
+  exists = bolt_store_has_journal (tt->store, "acl", "log");
+  g_assert_false (exists);
+
+}
+
+static void
+test_store_upgrade (TestStore *tt, gconstpointer user_data)
+{
+  g_autoptr(BoltDevice) dev = NULL;
+  g_autoptr(GError) err = NULL;
+  g_autoptr(DIR) root = NULL;
+  char uid[] = "fbc83890-e9bf-45e5-a777-b3728490989c";
+  guint version;
+  gboolean up;
+  gboolean ok;
+
+  /* simulate a version 0 store, i.e. has some entries,
+   * but no 'version' file */
+
+  root = bolt_opendir (tt->path, &err);
+  g_assert_no_error (err);
+  g_assert_nonnull (root);
+
+  version = bolt_store_get_version (tt->store);
+  g_assert_cmpuint (version, ==, BOLT_STORE_VERSION);
+
+  dev = g_object_new (BOLT_TYPE_DEVICE,
+                      "uid", uid,
+                      "name", "Laptop",
+                      "vendor", "GNOME.org",
+                      "status", BOLT_STATUS_DISCONNECTED,
+                      NULL);
+
+  ok = bolt_store_put_device (tt->store, dev, BOLT_POLICY_AUTO, NULL, &err);
+  g_assert_no_error (err);
+  g_assert_true (ok);
+
+  /* close the store, delete 'version' */
+  g_clear_object (&tt->store);
+
+  ok = bolt_unlink_at (dirfd (root), "version", 0, &err);
+  g_assert_no_error (err);
+  g_assert_true (ok);
+
+  /* re-create the store object */
+  tt->store = bolt_store_new (tt->path, &err);
+  g_assert_no_error (err);
+  g_assert_nonnull (tt->store);
+
+  version = bolt_store_get_version (tt->store);
+  g_assert_cmpuint (version, ==, 0);
+
+  /* no upgrade the store */
+  ok = bolt_store_upgrade (tt->store, &up, &err);
+  g_assert_no_error (err);
+  g_assert_true (ok);
+  g_assert_true (up);
+
+  /* assert the upgrade changed the version */
+  version = bolt_store_get_version (tt->store);
+  g_assert_cmpuint (version, ==, BOLT_STORE_VERSION);
+
+  /* upgrade again, check it did not do anything */
+  ok = bolt_store_upgrade (tt->store, &up, &err);
+  g_assert_no_error (err);
+  g_assert_true (ok);
+  g_assert_false (up);
+
+  version = bolt_store_get_version (tt->store);
+  g_assert_cmpuint (version, ==, BOLT_STORE_VERSION);
+
+  /* ensure 'upgrade' argument is optional */
+  ok = bolt_store_upgrade (tt->store, NULL, &err);
+  g_assert_no_error (err);
+  g_assert_true (ok);
+
 }
 
 int
 main (int argc, char **argv)
 {
+  TestContext test_context;
 
   setlocale (LC_ALL, "");
 
   g_test_init (&argc, &argv, NULL);
 
+  g_log_set_writer_func (test_context_logger, &test_context, NULL);
+
   bolt_dbus_ensure_resources ();
 
   g_test_add ("/daemon/key",
               TestStore,
-              NULL,
+              &test_context,
               test_store_setup,
               test_key,
               test_store_tear_down);
 
   g_test_add ("/daemon/store/basic",
               TestStore,
-              NULL,
+              &test_context,
               test_store_setup,
               test_store_basic,
               test_store_tear_down);
 
+  g_test_add ("/daemon/store/update",
+              TestStore,
+              &test_context,
+              test_store_setup,
+              test_store_update,
+              test_store_tear_down);
+
   g_test_add ("/daemon/store/config",
               TestStore,
-              NULL,
+              &test_context,
               test_store_setup,
               test_store_config,
               test_store_tear_down);
 
   g_test_add ("/daemon/store/invalid_data",
               TestStore,
-              NULL,
+              &test_context,
               test_store_setup,
               test_store_invalid_data,
               test_store_tear_down);
 
   g_test_add ("/daemon/store/times",
               TestStore,
-              NULL,
+              &test_context,
               test_store_setup,
               test_store_times,
               test_store_tear_down);
 
   g_test_add ("/daemon/store/domain",
               TestStore,
-              NULL,
+              &test_context,
               test_store_setup,
               test_store_domain,
               test_store_tear_down);
 
+  g_test_add ("/daemon/store/journal",
+              TestStore,
+              &test_context,
+              test_store_setup,
+              test_store_journal,
+              test_store_tear_down);
+
+  g_test_add ("/daemon/store/upgrade",
+              TestStore,
+              &test_context,
+              test_store_setup,
+              test_store_upgrade,
+              test_store_tear_down);
+
   return g_test_run ();
 }
