@@ -21,23 +21,20 @@
 #include "config.h"
 
 #include "bolt-unix.h"
-
 #include "bolt-io.h"
 #include "bolt-macros.h"
+#include "bolt-names.h"
 #include "bolt-str.h"
 #include "bolt-test.h"
 
-#include <glib-unix.h>
 #include <gio/gio.h>
 
-#include <errno.h>
 #include <locale.h>
 #include <sys/types.h>
 #include <sys/socket.h>
-#include <sys/stat.h>
 #include <sys/un.h>
-#include <sys/wait.h>
-#include <unistd.h> /* unlinkat, fork */
+#include <sys/wait.h> /* waitpid */
+#include <unistd.h>   /* fork */
 
 typedef struct TestDummy
 {
@@ -74,159 +71,19 @@ test_pid_is_alive (TestDummy *tt, gconstpointer user_data)
 
 typedef struct TestNotify
 {
-  BoltTmpDir tmpdir;
-  char      *socket_path;
-  guint      socket_watch;
-  int        socket_fd;
-
-  /* */
-  guint  counter;
-  GQueue messages;
+  NotifySocket *ns;
 } TestNotify;
 
-union ctrlmsg
-{
-  struct cmsghdr hdr;
-  guint8         buf[CMSG_SPACE (sizeof (struct ucred))];
-};
-
-static char *
-test_notify_revmsg (TestNotify *tt, gboolean queue)
-{
-  char data[4096];
-  char *msg;
-  struct iovec iov = {
-    .iov_base = data,
-    .iov_len = sizeof (data) - 1,
-  };
-  union ctrlmsg crtl = {};
-  struct msghdr hdr = {
-    .msg_iov = &iov,
-    .msg_iovlen = 1,
-    .msg_control = &crtl,
-    .msg_controllen = sizeof (crtl),
-  };
-  struct ucred *ucred = NULL;
-  ssize_t r;
-
-  /* MSG_TRUNC: return the real size */
-  r = recvmsg (tt->socket_fd, &hdr, MSG_DONTWAIT | MSG_CMSG_CLOEXEC | MSG_TRUNC);
-
-  if (r < 0)
-    {
-      if (errno == EINTR || errno == EAGAIN)
-        return NULL;
-
-      g_critical ("i/o error reading from notify socket: %m");
-      return NULL;
-    }
-
-  if (hdr.msg_flags & MSG_TRUNC || ((size_t) r > sizeof (data) - 1))
-    {
-      g_warning ("notification message truncated");
-      return NULL;
-    }
-
-  g_assert_cmpint (r, <, sizeof (data));
-
-  data[r] = '\0';
-
-  tt->counter++;
-  msg = g_strdup (data);
-
-  for (struct cmsghdr *c = CMSG_FIRSTHDR (&hdr);
-       c != NULL;
-       c = CMSG_NXTHDR (&hdr, c))
-    {
-      if (c->cmsg_level != SOL_SOCKET)
-        continue;
-      if (c->cmsg_type == SCM_CREDENTIALS &&
-          c->cmsg_len == CMSG_LEN (sizeof (struct ucred)))
-        ucred = (struct ucred *) (void *) CMSG_DATA (c);
-    }
-
-  if (queue)
-    g_queue_push_tail (&tt->messages, msg);
-
-  g_debug ("got message: '%s' [%s]", msg, bolt_yesno (queue));
-  if (ucred != NULL)
-    g_debug ("  ucred, pid: %i, uid: %li, gid: %li",
-             (int) ucred->pid, (long) ucred->uid, (long) ucred->gid);
-
-  return msg;
-}
-
-static gboolean
-got_notification (gpointer user_data)
-{
-  TestNotify *tt = (TestNotify *) user_data;
-
-  test_notify_revmsg (tt, TRUE);
-
-  return TRUE;
-}
-
 static void
 test_notify_setup (TestNotify *tt, gconstpointer data)
 {
-  g_autoptr(GError) err = NULL;
-  g_autoptr(GSource) source = NULL;
-  bolt_autoclose int fd = -1;
-  static const int one = 1;
-  struct sockaddr_un sau = {AF_UNIX, {'\0', }};
-  size_t socklen;
-  int r;
-
-  tt->tmpdir = bolt_tmp_dir_make ("bolt.unix.XXXXXX", &err);
-  g_assert_no_error (err);
-  g_assert_nonnull (tt->tmpdir);
-
-  fd = socket (AF_UNIX, SOCK_DGRAM | SOCK_CLOEXEC | SOCK_NONBLOCK, 0);
-  g_assert_cmpint (fd, >, -1);
-
-  tt->socket_path = g_build_filename (tt->tmpdir, "notify_socket", NULL);
-
-  strncpy (sau.sun_path, tt->socket_path, sizeof (sau.sun_path) - 1);
-
-  socklen =
-    offsetof (struct sockaddr_un, sun_path)
-    + strlen (sau.sun_path)
-    + 1;
-
-  r = bind (fd, &sau, socklen);
-  g_assert_cmpint (r, >, -1);
-
-  r = setsockopt (fd, SOL_SOCKET, SO_PASSCRED, &one, sizeof (one));
-  g_assert_cmpint (r, >, -1);
-
-  source = g_unix_fd_source_new (fd, G_IO_IN);
-  g_assert_nonnull (source);
-
-  g_source_set_callback (source, got_notification, tt, NULL);
-  tt->socket_watch = g_source_attach (source, NULL);
-
-  tt->socket_fd = bolt_steal (&fd, -1);
-  g_queue_init (&tt->messages);
-
-  g_debug ("notification socket at '%s'", sau.sun_path);
+  tt->ns = notify_socket_new ();
 }
 
 static void
 test_notify_teardown (TestNotify *tt, gconstpointer user)
 {
-  g_autoptr(GError) err = NULL;
-
-  g_clear_handle_id (&tt->socket_watch, g_source_remove);
-
-  if (tt->socket_fd > -1)
-    {
-      bolt_close (tt->socket_fd, &err);
-      g_assert_no_error (err);
-      tt->socket_fd = -1;
-    }
-
-  g_clear_pointer (&tt->tmpdir, bolt_tmp_dir_destroy);
-  g_queue_free_full (&tt->messages, g_free);
+  g_clear_pointer (&tt->ns, notify_socket_free);
 }
 
 static void
@@ -249,7 +106,7 @@ test_sd_notify (TestNotify *tt, gconstpointer user)
   g_assert_false (sent);
 
   /* invalid/unsupported destinations */
-  g_setenv ("NOTIFY_SOCKET", "INVALID SOCKET", TRUE);
+  g_setenv (BOLT_SD_NOTIFY_SOCKET, "INVALID SOCKET", TRUE);
   ok = bolt_sd_notify_literal (ref, &sent, &err);
   g_assert_error (err, G_IO_ERROR, G_IO_ERROR_NOT_SUPPORTED);
   g_assert_false (ok);
@@ -261,7 +118,7 @@ test_sd_notify (TestNotify *tt, gconstpointer user)
   g_assert_cmpuint (l, <, 1024);
   verylong = g_strnfill (l, 'a');
 
-  g_setenv ("NOTIFY_SOCKET", verylong, TRUE);
+  g_setenv (BOLT_SD_NOTIFY_SOCKET, verylong, TRUE);
   ok = bolt_sd_notify_literal (ref, &sent, &err);
   g_assert_error (err, G_IO_ERROR, G_IO_ERROR_INVALID_ARGUMENT);
   g_assert_false (ok);
@@ -269,7 +126,7 @@ test_sd_notify (TestNotify *tt, gconstpointer user)
   g_clear_error (&err);
 
   /* peer does not exist */
-  g_setenv ("NOTIFY_SOCKET", "@NONEXISTANTABSTRACT", TRUE);
+  g_setenv (BOLT_SD_NOTIFY_SOCKET, "@NONEXISTANTABSTRACT", TRUE);
   ok = bolt_sd_notify_literal (ref, &sent, &err);
   g_assert_error (err, G_IO_ERROR, G_IO_ERROR_CONNECTION_REFUSED);
   g_assert_false (ok);
@@ -277,19 +134,56 @@ test_sd_notify (TestNotify *tt, gconstpointer user)
   g_clear_error (&err);
 
   /* finally the VALID socket */
-  g_setenv ("NOTIFY_SOCKET", tt->socket_path, TRUE);
+  notify_socket_set_environment (tt->ns);
 
   ok = bolt_sd_notify_literal (ref, &sent, &err);
   g_assert_no_error (err);
   g_assert_true (ok);
   g_assert_true (sent);
 
-  msg = test_notify_revmsg (tt, FALSE);
+  msg = notify_socket_revmsg (tt->ns, FALSE);
   g_assert_nonnull (msg);
   g_assert_cmpstr (msg, ==, ref);
   g_free (msg);
 }
 
+static void
+test_sd_watchdog_enabled (TestDummy *tt, gconstpointer user_data)
+{
+  g_autoptr(GError) err = NULL;
+  g_autofree char *tmp = NULL;
+  guint64 val;
+  int r;
+
+  /* no env variable */
+  g_assert_null (g_getenv (BOLT_SD_WATCHDOG_USEC));
+
+  r = bolt_sd_watchdog_enabled (&val, &err);
+  g_assert_no_error (err);
+  g_assert_cmpint (r, ==, 0);
+
+  /* empty env variable [error] */
+  g_setenv (BOLT_SD_WATCHDOG_USEC, "", TRUE);
+  r = bolt_sd_watchdog_enabled (&val, &err);
+  g_assert_nonnull (err);
+  g_assert_cmpint (r, <, 0);
+  g_clear_error (&err);
+
+  /* invalid env variable [error] */
+  g_setenv (BOLT_SD_WATCHDOG_USEC, "NOT-A-NUMBER", TRUE);
+  r = bolt_sd_watchdog_enabled (&val, &err);
+  g_assert_nonnull (err);
+  g_assert_cmpint (r, <, 0);
+  g_clear_error (&err);
+
+  /* valid number, finally */
+  tmp = g_strdup_printf ("%d", 42 * G_USEC_PER_SEC);
+  g_setenv (BOLT_SD_WATCHDOG_USEC, tmp, TRUE);
+  r = bolt_sd_watchdog_enabled (&val, &err);
+  g_assert_no_error (err);
+  g_assert_cmpint (r, >, 0);
+}
+
 int
 main (int argc, char **argv)
 {
@@ -311,5 +205,12 @@ main (int argc, char **argv)
               test_sd_notify,
               test_notify_teardown);
 
+  g_test_add ("/common/unix/sd_watchdog_enabled",
+              TestDummy,
+              NULL,
+              NULL,
+              test_sd_watchdog_enabled,
+              NULL);
+
   return g_test_run ();
 }
